package ByteCode;
import java.util.*;
import ClassFile.*;
import DataFlow.*;
import Misc.*;

public class ByteCode implements ClassFileConst
{
  protected infoMethod	 method;
  protected infoAttrCode attr;
  protected byte[]	 code;
  private   Vector	 insts;
  private   Vector	 excepts;
  public    LabelManager lm;

  //
  // Loading
  //
  public ByteCode(infoMethod m) throws Exception {
    method  = m;
    attr    = method.getAttrCode();
    code    = attr.code;
    lm	    = new LabelManager(this);
    insts   = new Vector();
    // parse bytecode and register labels
    Instruction inst = null;
    int ofs, ops, sz = code.length;
    for (ofs = 0, ops = 0; ofs < sz; ops++) {
      inst = Instruction.AutoLoad(this, ofs);
      ofs += inst.length;
      inst.pc = ops;
      insts.addElement(inst);
    }
    if ( ofs != code.length )
      throw new ByteCodeException("ByteCode.ByteCode", "opcode overrun");
    // register exception labels
    ExceptionHandler eh = null;
    sz = attr.pc_start.length;
    excepts = new Vector();
    for (int i = 0; i < sz; i++) {
      eh = new ExceptionHandler();
      eh.pc_start   = lm.RegistLabel(attr.pc_start[i]);
      eh.pc_end	    = lm.RegistLabel(attr.pc_end[i]);
      eh.pc_handler = lm.RegistLabel(attr.pc_handler[i]);
      eh.catch_type = attr.catch_type[i];
      eh.catch_var  = getClassVar(attr.catch_type[i]);
      excepts.addElement(eh);
    }
    // resolving labels
    lm.Resolve();
    // setup CodeModify
    addinsts   = new Vector();
    addexcepts = new Vector();
    addconsts  = new Vector();
    addmigra   = new Vector();
    maxl = attr.max_locals;
    maxs = attr.max_stack;
    consts = method.owner.constant_pool.length;
  }

  //
  // Saving
  //
  public void Commit() throws ClassFileException {
    int i, sz, ofs;
    // insert codes
    sz = addinsts.size();
    AddInsts ai;
    for (i = 0; i < sz; i++) {
      ai = (AddInsts)addinsts.elementAt(i);
      while ( !ai.insts.empty() ) {
	Instruction inst = (Instruction)ai.insts.pop();
	inst.SetOwner(this);
	insts.insertElementAt(inst, ai.at);
      }
    }
    // bytecode offset adjustment
    sz = size();
    do {
      for (i = 0, ofs = 0; i < sz; i++) {
	elementAt(i).Adjust(ofs);
	ofs += elementAt(i).length;
      }
    } while ( !lm.Adjust() );
    // generate new bytecode
    code = new byte[ofs];
    for (i = 0; i < sz; i++)
      elementAt(i).Save();
    // update Code attribute
    attr.max_locals = maxl;
    attr.max_stack  = maxs;
    attr.code = code;
    ExceptionHandler eh = null;
    sz = excepts.size();
    attr.pc_start   = new int[sz];
    attr.pc_end	    = new int[sz];
    attr.pc_handler = new int[sz];
    attr.catch_type = new infoCPindex[sz];
    for (i = 0; i < sz; i++) {
      eh = (ExceptionHandler)excepts.elementAt(i);
      attr.pc_start[i]   = eh.pc_start.offset;
      attr.pc_end[i]	 = eh.pc_end.offset;
      attr.pc_handler[i] = eh.pc_handler.offset;
      attr.catch_type[i] = eh.catch_type;
    }
    // adjust code attribute
    attr.Adjust();
    // change descriptor
    sz = addmigra.size();
    for (i = 0; i < sz; i++) {
      infoCPindex cpi = (infoCPindex)addmigra.elementAt(i); // methodref
      cpi = method.owner.constant_pool[cpi.index].getIndex(1); // nameandtype
      cpi.cp.setDesc(getMigratory(cpi.cp.getIndex(1)));
    }
    method.desc = getMigratory(method.desc);
    // add constant pool
    method.owner.addConstantPool(addconsts);
  }

  //
  // DataFlow
  //
  public MySet CatchException(Instruction inst, State state) throws DataFlowException {
    MySet flows = new MySet();
    State stat = new State(state); stat.stack.Flash();
    int i, sz = excepts.size();
    for (i = 0; i < sz; i++) {
      ExceptionHandler eh = (ExceptionHandler)excepts.elementAt(i);
      if ( eh.pc_start.getInst().pc > inst.pc ||
	   eh.pc_end.getInst().pc <= inst.pc )
	continue;
      State cstat = new State(stat);
      cstat.stack.push(eh.catch_var);
      flows.add(new Flow(eh.pc_handler.getInst(), cstat));
    }
    return flows;
  }

  public Vector getInitialLocalVars() throws DataFlowException {
    String here = "ByteCode.getInitialLocalVars";
    try {
      String desc = method.desc.cp.getString();
      int start = desc.indexOf(DC_ARGS_BEGIN); 
      int end	= desc.lastIndexOf(DC_ARGS_END);
      Vector v;
      if ( end - start <= 1 )
	v = new Vector();
      else
	v = Var.getVars(desc.substring(start+1, end));
      if ( (method.flag & ACC_STATIC) == 0 ) {
	infoCPindex cpi = method.owner.this_class.cp.getIndex();
	String name = DC_CLASS_BEGIN + cpi.cp.getString() + DC_CLASS_END;
	v.insertElementAt(Var.getVar(name), 0);
      }
      return v;
    } catch (ClassFileException e) {
      throw new DataFlowException(here, null);
    }
  }

  protected Var getClassVar(int cindex) throws DataFlowException {
    String here = "ByteCode.getClassVar(int)";
    try {
      return getClassVar(method.owner.getCPindex(cindex, CONSTANT_Class));
    } catch (ClassFileException e) {
      throw new DataFlowException(here, null);
    }
  }

  protected Var getClassVar(infoCPindex cpi) throws DataFlowException {
    String here = "ByteCode.getClassVar(infoCPindex)";
    if ( cpi.index == 0 ) return Var.vthrow;
    try {
      cpi = cpi.cp.getIndex(); // CONSTANT_Utf8
      String name = cpi.cp.getString();
      if ( name.charAt(0) != DC_ARRAY )
	name = DC_CLASS_BEGIN + name + DC_CLASS_END;
      return Var.getVar(name);
    } catch (ClassFileException e) {
      throw new DataFlowException(here, null);
    }
  }

  protected Var getFieldVar(int cindex) throws DataFlowException {
    String here = "ByteCode.getFieldVar";
    try {
      infoCPindex cpi;
      cpi = method.owner.getCPindex(cindex, CONSTANT_Fieldref);
      cpi = cpi.cp.getIndex(1); // CONSTANT_NameAndType
      cpi = cpi.cp.getIndex(1); // CONSTANT_Utf8
      return Var.getVar(cpi.cp.getString());
    } catch (ClassFileException e) {
      throw new DataFlowException(here, null);
    }
  }

  protected Vector getMethodVars(int cindex) throws DataFlowException {
    String here = "ByteCode.getMethodVars";
    try {
      infoCPindex cpi;
      cpi = method.owner.getCPindex(cindex, CONSTANT_Methodref);
      cpi = cpi.cp.getIndex(1); // CONSTANT_NameAndType
      cpi = cpi.cp.getIndex(1); // CONSTANT_Utf8
      String desc = cpi.cp.getString();
      int start = desc.indexOf(DC_ARGS_BEGIN); 
      int end	= desc.lastIndexOf(DC_ARGS_END);
      Vector v;
      if ( end - start <= 1 )
	v = new Vector();
      else
	v = Var.getVars(desc.substring(start+1, end));
      v.insertElementAt(Var.getVar(desc.substring(end+1)), 0);
      return v;
    } catch (ClassFileException e) {
      throw new DataFlowException(here, null);
    }
  }

  protected Var getConstVar(int cindex) throws DataFlowException {
    String here = "ByteCode.getConstVar";
    try {
      infoCPindex cpi;
      cpi = method.owner.getCPindex(cindex, CONSTANT_Invalid);
      switch ( cpi.cp.tag ) {
      case CONSTANT_Integer: return Var.vint;
      case CONSTANT_Float:   return Var.vfloat;
      case CONSTANT_String:  return Var.vstring;
      case CONSTANT_Long:    return Var.vlong;
      case CONSTANT_Double:  return Var.vdouble;
      }
      throw new DataFlowException(here, "unknown atype");
    } catch (ClassFileException e) {
      throw new DataFlowException(here, null);
    }
  }

  //
  // code modifying
  //
  private Vector addinsts, addexcepts, addconsts, addmigra;
  private int maxs, maxl, consts;

  public int addClass(String clazz) {
    int iclazz = consts + addconsts.size();
    addconsts.addElement(infoCP.makeUtf8(clazz));
    addconsts.addElement(infoCP.makeClass(iclazz));
    return iclazz + 1;
  }

  public int addField(int iclazz, String name, String desc) {
    int iname = consts + addconsts.size();
    addconsts.addElement(infoCP.makeUtf8(name));
    addconsts.addElement(infoCP.makeUtf8(desc));
    addconsts.addElement(infoCP.makeNameAndType(iname, iname+1));
    addconsts.addElement(infoCP.makeFieldref(iclazz, iname+2));
    return iname+3;
  }

  public int addMethod(int iclazz, String name, String desc) {
    int iname = consts + addconsts.size();
    addconsts.addElement(infoCP.makeUtf8(name));
    addconsts.addElement(infoCP.makeUtf8(desc));
    addconsts.addElement(infoCP.makeNameAndType(iname, iname+1));
    addconsts.addElement(infoCP.makeMethodref(iclazz, iname+2));
    return iname+3;
  }

  public void addException(Label pc_start, Label pc_end, Label pc_catch, int cindex) {
    ExceptionHandler eh = new ExceptionHandler();
    eh.pc_start   = pc_start;
    eh.pc_end	  = pc_end;
    eh.pc_handler = pc_catch;
    eh.catch_type = new infoCPindex(cindex);
    eh.catch_var  = null;
    excepts.addElement(eh);
  }

  public infoCPindex getMigratory(infoCPindex cpi) throws ClassFileException {
    String desc = (cpi.cp != null) ? cpi.cp.getString()
		: ((infoCP)addconsts.elementAt(cpi.index-consts)).getString();
    if ( desc.startsWith("(Ljavago/StackFrame;") )
      return cpi;
    desc = "(Ljavago/StackFrame;" + desc.substring(1);
    addconsts.addElement(infoCP.makeUtf8(desc));
    return new infoCPindex(consts+addconsts.size()-1);
  }

  public void addMigratory(opINVOKE invoke) {
    addmigra.addElement(new infoCPindex(invoke.cindex));
  }

  public boolean isStatic() {
    return (method.flag & ACC_STATIC) != 0;
  }

  public Vector Survey() throws Exception {
    Hashtable hash = new Hashtable();
    Vector check = new Vector();
    elementAt(0).fSaveState = true;
    check.addElement(elementAt(0));
    int i, sz = size();
    for (i = 0; i < sz; i++) {
      Instruction inst = elementAt(i);
      if ( !(inst instanceof opINVOKE) )
	continue;
      String name = getMethodName(((opINVOKE)inst).cindex);
      if ( name.endsWith(">") )
	continue;
      Boolean bool = (Boolean)hash.get(name);
      if ( bool == null ) {
	System.out.print(name + " <y/n>? ");
	System.in.skip(System.in.available());
	if ( System.in.read() == 'y' )
	  bool = new Boolean(true);
	else
	  bool = new Boolean(false);
	hash.put(name, bool);
      }
      if ( bool.booleanValue() ) {
	inst.fSaveState = true;
	check.addElement(inst);
      }
    }
    return check;
  }

  public int getLastInstNumber() {
    return insts.size();
  }

  public void insertInsts(int at, Stack ins) {
    AddInsts ai = new AddInsts(at, ins);
    int i, sz = addinsts.size();
    for (i = 0; i < sz; i++) {
      AddInsts tmp = (AddInsts)addinsts.elementAt(i);
      if ( tmp.at < at )
	break;
    }
    addinsts.insertElementAt(ai, i);
  }

  protected String getMethodName(int cindex) throws ClassFileException {
    String here = "ByteCode.getMethodName";
    infoCPindex cpi, cpi0, cpi1;
    cpi = method.owner.getCPindex(cindex, CONSTANT_Methodref);
    cpi0 = cpi.cp.getIndex(0);	// CONSTANT_Class
    cpi0 = cpi0.cp.getIndex();	// CONSTANT_Utf8
    cpi1 = cpi.cp.getIndex(1);	// CONSTANT_NameAndType
    cpi1 = cpi1.cp.getIndex(0); // CONSTANT_Utf8
    return cpi0.cp.getString() + "/" + cpi1.cp.getString();
  }

  public Vector getMethodVars(opINVOKE invoke) {
    try {
      return getMethodVars(invoke.cindex);
    } catch (DataFlowException e) {
      e.Display();
      return null;
    }
  }

  //
  // utilities
  //
  public int getMaxLocals() { return attr.max_locals; }
  public int getMaxStack()  { return attr.max_stack; }
  public int setMaxLocals(int i) { return maxl = i; }
  public int setMaxStack(int i)  { return maxs = i; }

  public Instruction NextInst(Instruction inst)
    throws DataFlowException {
    if ( inst.pc+1 >= size() )
      throw new DataFlowException("ByteCode.NextInst", null);
    return elementAt(inst.pc+1);
  }

  public Instruction elementAt(int i) {
    return (Instruction)insts.elementAt(i);
  }

  public int size() {
    return insts.size();
  }

  //
  // debug
  //
  public String getName() {
    try {
      return method.name.cp.getString();
    } catch (Misc.MyException e) {
      e.Display();
      return e.toString();
    }
  }

  public String getClassName() {
    try {
      infoCPindex cpi = method.owner.this_class;
      cpi = cpi.cp.getIndex();
      return cpi.cp.getString();
    } catch (Misc.MyException e) {
      e.Display();
      return e.toString();
    }
  }
}

class ExceptionHandler
{
  protected Label pc_start, pc_end, pc_handler;
  protected infoCPindex catch_type;
  protected Var catch_var;
}

class AddInsts
{
  protected int at;
  protected Stack insts;
  protected AddInsts(int _at, Stack _insts) {
    at	  = _at;
    insts = _insts;
  }
}
