package ByteCode;
import java.util.*;
import Misc.*;

public class LabelManager
{
  private ByteCode bc;
  private Hashtable labels, subrs;
  private int uniq;

  public LabelManager(ByteCode owner) {
    bc = owner;
    labels = new Hashtable();
    subrs = new Hashtable();
    uniq = -1;
  }

  //
  // ByteCode
  //
  protected Label RegistLabel(int target) {
    Integer key = new Integer(target);
    if ( labels.containsKey(key) )
      return (Label)labels.get(key);
    Label label = new Label(target);
    labels.put(key, label);
    return label;
  }

  protected void RegistSubroutine(Label subr, Label retto) {
    MySet rettos = (MySet)subrs.get(subr);
    if ( rettos == null ) {
      rettos = new MySet();
      subrs.put(subr, rettos);
    }
    rettos.add(retto);
  }

  protected void Resolve() throws ByteCodeException {
    String here = "LabelManager.Resolve";
    int insts = bc.size();
    if ( insts == 0 )
      throw new ByteCodeException(here, "method has no instructions");
    int lookup[] = new int[insts];
    for (int i = 0; i < insts; i++)
      lookup[i] = bc.elementAt(i).offset;
    Enumeration enum = labels.elements();
    while ( enum.hasMoreElements() ) {
      Label label = (Label)enum.nextElement();
      int l = 0, h = insts-1, idx, ofs = label.offset, tmp;
      while ( true ) {
	if ( h - l < 4 ) {
	  for (idx = l; idx <= h; idx++)
	    if ( lookup[idx] == ofs )
	      break;
	  if ( idx > h )
	    throw new ByteCodeException(here, "target address " + ofs + " not found");
	  break;
	}
	idx = (l + h) / 2;
	tmp = lookup[idx];
	if ( tmp == ofs ) break;
	if ( tmp < ofs )  l = idx;
	else		  h = idx;
      }
      label.Resolve(bc.elementAt(idx));
    }
  }

  protected boolean Adjust() {
    boolean adjusted = true;
    Enumeration enum = labels.elements();
    while ( enum.hasMoreElements() )
      adjusted &= ((Label)enum.nextElement()).Adjust();
    return adjusted;
  }

  //
  // DataFlow
  //
  public void MarkCheckPoints() {
    Enumeration enum = labels.elements();
    while ( enum.hasMoreElements() )
      ((Label)enum.nextElement()).getInst().fSaveState = true;
  }

  public MySet getReturnAddresses(Label subr) {
    return (MySet)subrs.get(subr);
  }

  //
  // CodeModify
  //
  public Label RegistLabel(Instruction inst) {
    Label label;
    if ( inst.offset >= 0 )
      label = RegistLabel(inst.offset);
    else
      label = RegistLabel(uniq--);
    label.Resolve(inst);
    return label;
  }
}
