package ByteCode;
import DataFlow.*;
import Misc.*;

// IF[EQ,NE,LT,GE,GT,LE]
// IF_ICMP[EQ,NE,LT,GE,GT,LE]
// IF_ACMP[EQ,NE]

public class opCOND extends Instruction
{
  Label target;

  public opCOND(ByteCode bc, int ofs) {
    super(bc, ofs);
    length = 3;
    target = bc.lm.RegistLabel(offset + ReadShort(ofs+1));
  }

  public void Adjust(int ofs) {
    super.Adjust(ofs);
    if ( InShortRange(target) )
      length = 3;
    else
      length = 8; // IF* + GOTO_W
  }

  public void Save() {
    if ( length == 3 ) {
      super.Save();
      WriteShort(offset+1, RelShort(target));
    } else {
      int opposite = (((opcode-IFEQ)&1) == 0) ? opcode+1 : opcode-1;
      WriteByteU(offset, opposite);
      WriteShort(offset+1, 3);
      WriteByteU(offset+3, GOTO_W);
      WriteInt(offset+4, target.offset-(offset+3));
    }
  }

  public NextState Analyze(State stat) throws DataFlowException {
    if ( SaveState(stat) )
      return null;
    switch ( opcode ) {
    case IFEQ: case IFNE: case IFLT:
    case IFGE: case IFGT: case IFLE:
      stat.stack.pop(T_INT);
      break;
    case IF_ICMPEQ: case IF_ICMPNE: case IF_ICMPLT:
    case IF_ICMPGE: case IF_ICMPGT: case IF_ICMPLE:
      stat.stack.pop(T_INT);
      stat.stack.pop(T_INT);
      break;
    case IF_ACMPEQ: case IF_ACMPNE:
      stat.stack.pop(T_REFER);
    case IFNULL:    case IFNONNULL:
      stat.stack.pop(T_REFER);
      break;
    default:
      throw new DataFlowException("opCOND", strOpcode);
    }
    NextState ns = new NextState(stat);
    ns.push(new Flow(bc.NextInst(this), stat));
    ns.push(new Flow(target.getInst(), stat));
    return ns;
  }

  public opCOND(int op, Label to) {
    super(op);
    target = to;
  }
}
