package ByteCode;
import DataFlow.*;
import Misc.*;

// LOOKUPSWITCH

public class opLOOKUP extends Instruction
{
  Label target, targets[];
  int npairs, lookups[];

  public opLOOKUP(ByteCode bc, int ofs) throws ByteCodeException {
    super(bc, ofs);
    int pad = (ofs + 4) & 0xFFFFFFFC;
    target  = bc.lm.RegistLabel(offset + ReadInt(pad));
    npairs  = ReadInt(pad+4);
    length  = pad + 8 + 8*npairs - offset;
    if ( offset+length >= bc.code.length )
      throw new ByteCodeException("opLOOKUP.opLOOKUP", "illegal opcode (buffer overrun)");
    lookups = new int[npairs];
    targets = new Label[npairs];
    for (int i = 0; i < npairs; i++) {
      lookups[i] = ReadInt(pad+8*(i+1));
      targets[i] = bc.lm.RegistLabel(offset + ReadInt(pad+8*(i+1)+4));
    }
  }

  public void Adjust(int ofs) {
    super.Adjust(ofs);
    int pad = (ofs + 4) & 0xFFFFFFFC;
    length = pad + 8 + 8*npairs - offset;
  }

  public void Save() {
    super.Save();
    int pad = (offset + 4) & 0xFFFFFFFC;
    for (int i = offset+1; i < pad; i++)
      WriteByteU(i, 0);
    WriteInt(pad, RelInt(target));
    WriteInt(pad+4, npairs);
    for (int i = 0; i < npairs; i++) {
      WriteInt(pad+8*(i+1), lookups[i]);
      WriteInt(pad+8*(i+1)+4, RelInt(targets[i]));
    }
  }

  public NextState Analyze(State stat) throws DataFlowException {
    if ( SaveState(stat) )
      return null;
    stat.stack.pop(T_INT);
    NextState ns = new NextState(stat);
    ns.push(new Flow(target.getInst(), stat));
    for (int i = 0; i < npairs; i++)
      ns.push(new Flow(targets[i].getInst(), stat));
    return ns;
  }
}
