package ByteCode;
import DataFlow.*;
import Misc.*;

// TABLESWITCH

public class opTABLE extends Instruction
{
  Label target, targets[];
  int low, high;

  public opTABLE(ByteCode bc, int ofs) throws ByteCodeException {
    super(bc, ofs);
    String here = "opTABLE.opTABLE";
    int pad = (ofs + 4) & 0xFFFFFFFC;
    target = bc.lm.RegistLabel(offset + ReadInt(pad));
    low    = ReadInt(pad+4);
    high   = ReadInt(pad+8);
    if ( low > high )
      throw new ByteCodeException(here, "illegal opcode (low > high)");
    int n  = high - low + 1;
    length = pad + 12 + 4 * n - offset;
    if ( offset+length >= bc.code.length )
      throw new ByteCodeException(here, "illegal opcode (buffer overrun)");
    targets = new Label[n];
    for (int i = 0; i < n; i++)
      targets[i] = bc.lm.RegistLabel(offset + ReadInt(pad+4*(i+3)));
  }

  public void Adjust(int ofs) {
    super.Adjust(ofs);
    int pad = (ofs + 4) & 0xFFFFFFFC;
    length = pad + 12 + 4 * (high-low+1) - offset;
  }

  public void Save() {
    super.Save();
    int pad = (offset + 4) & 0xFFFFFFFC;
    for (int i = offset+1; i < pad; i++)
      WriteByteU(i, 0);
    WriteInt(pad, RelInt(target));
    WriteInt(pad+4, low);
    WriteInt(pad+8, high);
    int n = high - low + 1;
    for (int i = 0; i < n; i++)
      WriteInt(pad+4*(i+3), RelInt(targets[i]));
  }

  public NextState Analyze(State stat) throws DataFlowException {
    if ( SaveState(stat) )
      return null;
    stat.stack.pop(T_INT);
    NextState ns = new NextState(stat);
    ns.push(new Flow(target.getInst(), stat));
    for (int i = 0; i < targets.length; i++)
      ns.push(new Flow(targets[i].getInst(), stat));
    return ns;
  }

  public opTABLE(int op, int _low, int _high, Label _target, Label[] _targets) {
    super(op);
    low = _low;
    high = _high;
    target = _target;
    targets = _targets;
    //Adjust();
  }
}
