package ByteCode;
import DataFlow.*;

// WIDE
// - ILOAD,LLOAD,FLOAD,DLOAD,ALOAD
// - ISTORE,LSTORE,FSTORE,DSTORE,ASTORE
// - IINC

public class opWIDE extends Instruction
{
  int subopcode, lindex, amount;

  public opWIDE(ByteCode bc, int ofs) {
    super(bc, ofs);
    subopcode = ReadByteU(ofs+1);
    lindex = ReadShort(ofs+2);
    switch ( subopcode ) {
    case IINC:
      length = 6;
      amount = ReadShort(ofs+4);
      break;
    default:
      length = 4;
      break;
    }
  }

  public void Save() {
    super.Save();
    WriteByteU(offset+1, subopcode);
    WriteShort(offset+2, lindex);
    if ( subopcode == IINC )
      WriteShort(offset+4, amount);
  }

  public NextState Analyze(State stat) throws DataFlowException {
    if ( SaveState(stat) )
      return null;
    switch ( subopcode ) {
    case IINC:   stat.local.Read(lindex, T_INT); break;
    case ILOAD:  stat.stack.push(stat.local.Read(lindex, T_INT)); break;
    case LLOAD:  stat.stack.push(stat.local.Read(lindex, T_LONG)); break;
    case FLOAD:  stat.stack.push(stat.local.Read(lindex, T_FLOAT)); break;
    case DLOAD:  stat.stack.push(stat.local.Read(lindex, T_DOUBLE)); break;
    case ALOAD:	 stat.stack.push(stat.local.Read(lindex, T_RETURN)); break;
    case ISTORE: stat.local.Write(lindex, stat.stack.pop(T_INT)); break;
    case LSTORE: stat.local.Write(lindex, stat.stack.pop(T_LONG)); break;
    case FSTORE: stat.local.Write(lindex, stat.stack.pop(T_FLOAT)); break;
    case DSTORE: stat.local.Write(lindex, stat.stack.pop(T_DOUBLE)); break;
    case ASTORE: stat.local.Write(lindex, stat.stack.pop(T_RETURN)); break;
    default:	 throw new DataFlowException("opWIDE", strOpcode);
    }
    NextState ns = new NextState(stat);
    ns.push(new Flow(bc.NextInst(this), stat));
    return ns;
  }
}
