package ClassFile;
import java.io.*;
import java.util.*;

public class ClassFile implements ClassFileConst
{
  public int		magic;
  public int		minor_version, major_version;
  public infoCP		constant_pool[];
  public int		access_flag;
  public infoCPindex	this_class, super_class;
  public infoCPindex	interfaces[];
  public infoField	fields[];
  public infoMethod	methods[];
  public infoAttribute	attributes[];
  public boolean	isExtra;

  public ClassFile(DataInputStream in) throws Exception {
    Load(in);
  }

  public void Load(DataInputStream in) throws Exception {
    String here = "ClassFile.Load";
    // magic, version
    magic = in.readInt();
    if ( magic != JAVA_MAGIC )
      throw new ClassFileException(here, "unexpected magic number");
    minor_version = in.readUnsignedShort();
    major_version = in.readUnsignedShort();
    // Constant Pool
    constant_pool = new infoCP[in.readUnsignedShort()];
    if ( constant_pool.length > 0 )
      constant_pool[0] = infoCP.Invalid;
    for (int i = 1; i < constant_pool.length; i++) {
      constant_pool[i] = new infoCP(in);
      if ( constant_pool[i].isDWORD )
	if ( ++i < constant_pool.length )
	  constant_pool[i] = infoCP.Invalid;
	else
	  throw new ClassFileException(here, "constant pool overrun");
    }
    for (int i = 1; i < constant_pool.length; i++)
      constant_pool[i].Resolve(constant_pool);
    // access flag
    access_flag = in.readUnsignedShort();
    this_class = new infoCPindex(in, constant_pool, CONSTANT_Class);
    super_class = new infoCPindex(in, constant_pool, CONSTANT_Class);
    // interfaces
    interfaces = new infoCPindex[in.readUnsignedShort()];
    for (int i = 0; i < interfaces.length; i++)
      interfaces[i] = new infoCPindex(in, constant_pool, CONSTANT_Class);
    // fields
    fields = new infoField[in.readUnsignedShort()];
    for (int i = 0; i < fields.length; i++)
      fields[i] = new infoField(this, in);
    // methods
    methods = new infoMethod[in.readUnsignedShort()];
    for (int i = 0; i < methods.length; i++)
      methods[i] = new infoMethod(this, in);
    // attributes
    attributes = new infoAttribute[in.readUnsignedShort()];
    for (int i = 0; i < attributes.length; i++)
      attributes[i] = infoAttribute.AutoLoad(this, in);
    isExtra = (in.read() != -1);
  }

  public void Save(DataOutputStream out) throws Exception {
    // magic, version
    out.writeInt(magic);
    out.writeShort(minor_version);
    out.writeShort(major_version);
    // Constant Pool
    out.writeShort(constant_pool.length);
    for (int i = 1; i < constant_pool.length; i++) {
      constant_pool[i].Save(out);
      if ( constant_pool[i].isDWORD ) i++;
    }
    // access flag
    out.writeShort(access_flag);
    this_class.Save(out);
    super_class.Save(out);
    // interfaces
    out.writeShort(interfaces.length);
    for (int i = 0; i < interfaces.length; i++)
      interfaces[i].Save(out);
    // fields
    out.writeShort(fields.length);
    for (int i = 0; i < fields.length; i++)
      fields[i].Save(out);
    // methods
    out.writeShort(methods.length);
    for (int i = 0; i < methods.length; i++)
      methods[i].Save(out);
    // attributes
    out.writeShort(attributes.length);
    for (int i = 0; i < attributes.length; i++)
      attributes[i].Save(out);
  }

  //
  // CodeModify
  //
  public void addConstantPool(Vector consts) {
    int orgsz = constant_pool.length, addsz = consts.size();
    infoCP[] newCP = new infoCP[orgsz+addsz];
    for (int i = 0; i < orgsz; i++)
      newCP[i] = constant_pool[i];
    for (int i = 0; i < addsz; i++)
      newCP[orgsz+i] = (infoCP)consts.elementAt(i);
    constant_pool = newCP;
  }

  //
  // utilities
  //
  public infoCPindex getCPindex(int index, int tag) throws ClassFileException {
    String here = "ClassFile.getCPindex";
    if ( constant_pool.length <= index )
      throw new ClassFileException(here, null);
    if ( tag == CONSTANT_Invalid )
      tag = constant_pool[index].tag;
    if ( constant_pool[index].tag != tag )
      throw new ClassFileException(here, null);
    return new infoCPindex(index, constant_pool, tag);
  }
}
