package ClassFile;
import java.io.*;

public class infoCP implements ClassFileConst
{
  static public infoCP Invalid = new infoCP();

  public int	 tag;
  public boolean isDWORD;
  public Object  value;

  public infoCP() {
    tag = CONSTANT_Invalid;
  }

  public infoCP(DataInputStream in) throws Exception {
    Load(in);
  }

  public void Load(DataInputStream in) throws Exception {
    tag = in.readUnsignedByte();
    isDWORD = false;
    switch ( tag ) {
    case CONSTANT_Utf8:
      value = in.readUTF();
      break;
    case CONSTANT_Integer:
      value = new Integer(in.readInt());
      break;
    case CONSTANT_Float:
      value = new Float(in.readFloat());
      break;
    case CONSTANT_Long:
      isDWORD = true;
      value = new Long(in.readLong());
      break;
    case CONSTANT_Double:
      isDWORD = true;
      value = new Double(in.readDouble());
      break;
    case CONSTANT_Class:
    case CONSTANT_String:
      value = new infoCPindex(in);
      break;
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_NameAndType:
      value = new infoCPindex[2];
      ((infoCPindex[])value)[0] = new infoCPindex(in);
      ((infoCPindex[])value)[1] = new infoCPindex(in);
      break;
    default:
      throw new ClassFileException("infoCP.Load", "unexpected CONSTANT_Tag");
    }
  }

  public void Save(DataOutputStream out) throws Exception {
    out.writeByte(tag);
    switch ( tag ) {
    case CONSTANT_Utf8:
      out.writeUTF((String)value);
      break;
    case CONSTANT_Integer:
      out.writeInt(((Integer)value).intValue());
      break;
    case CONSTANT_Float:
      out.writeFloat(((Float)value).floatValue());
      break;
    case CONSTANT_Long:
      out.writeLong(((Long)value).longValue());
      break;
    case CONSTANT_Double:
      out.writeDouble(((Double)value).doubleValue());
      break;
    case CONSTANT_Class:
    case CONSTANT_String:
      ((infoCPindex)value).Save(out);
      break;
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_NameAndType:
      ((infoCPindex[])value)[0].Save(out);
      ((infoCPindex[])value)[1].Save(out);
      break;
    default:
      throw new ClassFileException("infoCP.Save", "unexpected CONSTANT_Tag");
    }
  }

  public void Resolve(infoCP[] CPs) throws ClassFileException {
    switch ( tag ) {
    case CONSTANT_Class:
    case CONSTANT_String:
      ((infoCPindex)value).Resolve(CPs, CONSTANT_Utf8);
      break;
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
      ((infoCPindex[])value)[0].Resolve(CPs, CONSTANT_Class);
      ((infoCPindex[])value)[1].Resolve(CPs, CONSTANT_NameAndType);
      break;
    case CONSTANT_NameAndType:
      ((infoCPindex[])value)[0].Resolve(CPs, CONSTANT_Utf8);
      ((infoCPindex[])value)[1].Resolve(CPs, CONSTANT_Utf8);
      break;
    case CONSTANT_Utf8:
    case CONSTANT_Integer:
    case CONSTANT_Float:
    case CONSTANT_Long:
    case CONSTANT_Double:
      break;
    default:
      throw new ClassFileException("infoCP.Resolve", "unexpected CONSTANT_Tag");
    }
  }

  //
  // CodeModify
  //
  private infoCP(String str) {
    tag = CONSTANT_Utf8;
    value = str;
  }

  private infoCP(int _tag, int cpi) {
    tag = _tag;
    value = new infoCPindex(cpi);
  }

  private infoCP(int _tag, int cp0, int cp1) {
    tag = _tag;
    value = new infoCPindex[2];
    ((infoCPindex[])value)[0] = new infoCPindex(cp0);
    ((infoCPindex[])value)[1] = new infoCPindex(cp1);
  }

  public void setDesc(infoCPindex cpi) {
    ((infoCPindex[])value)[1] = cpi;
  }

  public static infoCP makeUtf8(String str) {
    return new infoCP(str);
  }

  public static infoCP makeClass(int name) {
    return new infoCP(CONSTANT_Class, name);
  }

  public static infoCP makeNameAndType(int name, int type) {
    return new infoCP(CONSTANT_NameAndType, name, type);
  }

  public static infoCP makeFieldref(int clazz, int nandt) {
    return new infoCP(CONSTANT_Fieldref, clazz, nandt);
  }

  public static infoCP makeMethodref(int clazz, int nandt) {
    return new infoCP(CONSTANT_Methodref, clazz, nandt);
  }

  //
  // utilities
  //
  public String getString() throws ClassFileException {
    if ( tag == CONSTANT_Utf8 )
      return (String)value;
    throw new ClassFileException(null, null);
  }

  public infoCPindex getIndex() throws ClassFileException {
    switch ( tag ) {
    case CONSTANT_Class:
    case CONSTANT_String:
      return (infoCPindex)value;
    }
    throw new ClassFileException(null, null);
  }

  public infoCPindex getIndex(int i) throws ClassFileException {
    switch ( tag ) {
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_NameAndType:
      return ((infoCPindex[])value)[i];
    }
    throw new ClassFileException(null, null);
  }
}
