package CodeModify;
import java.util.*;
import ByteCode.*;
import ClassFile.*;
import DataFlow.*;

// Ճ[h!!!

public class CodeModify implements ByteCodeConst
{
  private boolean fDebug;
  private ByteCode bc;
  private Vector modify;

  public CodeModify(ByteCode bytecode, boolean fdebug) {
    bc = bytecode;
    fDebug = fdebug;
  }

  public boolean Survey() throws CodeModifyException {
    String here = "CodeModify.Survey";
    String name = bc.getName();
    if ( name.charAt(0) == '<' )
      return false;
    System.out.println("Starting: " + name);
    try {
      modify = bc.Survey();
    } catch (ClassFileException e) {
      e.Display();
      throw new CodeModifyException(here, null);
    } catch (Exception e) {
      throw new CodeModifyException(here, "I/O Error (user input)");
    }
    return true;
  }

  private int maxl, maxs, nmaxl, nmaxs, lmps, lms, lnotify, ltmp, lstack;
  private Label la_resume[], la_start[], la_end[], la_catch[];
  private int cpStackFrame, cpStackFrameChild;
  private int cpNotifyGone, cpNotifyGoneBottomofStack;
  private int cpMyState, cpMyStateInit, cpMyStateMEntryPoint;
  private int cpArrayI, cpArrayL, cpArrayF, cpArrayD, cpArrayA, cpArray[];

  public void Modify() throws CodeModifyException {
    // prepairing
    maxl    = bc.getMaxLocals();
    maxs    = bc.getMaxStack();
    nmaxl   = maxl + 10 + maxs*2;
    nmaxs   = (maxs < 4) ? 4 : maxs;
    lmps    = maxl;
    lms	    = maxl+1;
    lnotify = maxl+2;
    ltmp    = maxl+3;
    lstack  = maxl+8;
    int invokes = modify.size()-1;
    la_resume = new Label[invokes];
    la_start  = new Label[invokes];
    la_end    = new Label[invokes];
    la_catch  = new Label[invokes];
    // constants
    cpStackFrame = bc.addClass("javago/StackFrame");
    cpStackFrameChild = bc.addField(cpStackFrame, "Child", "Ljavago/StackFrame;");
    cpNotifyGone = bc.addClass("javago/NotifyGone");
    cpNotifyGoneBottomofStack = bc.addField(cpNotifyGone, "BottomofStack", "Ljavago/StackFrame;");
    {
      String clazz = bc.getClassName();
      int idx = clazz.lastIndexOf('/');
      if ( idx >= 0 )
	cpMyState = bc.addClass(clazz.substring(0, idx+1) + "ST"
				+ clazz.substring(idx+1) + bc.getName());
      else
	cpMyState = bc.addClass("ST" + clazz + bc.getName());
    }
    cpMyStateInit = bc.addMethod(cpMyState, "<init>", "()V");
    cpMyStateMEntryPoint = bc.addField(cpMyState, "M_EntryPoint", "I");
    cpArray = new int[5];
    cpArray[0] = bc.addField(cpMyState, "ArrayI", "[I");
    cpArray[1] = bc.addField(cpMyState, "ArrayL", "[J");
    cpArray[2] = bc.addField(cpMyState, "ArrayF", "[F");
    cpArray[3] = bc.addField(cpMyState, "ArrayD", "[D");
    cpArray[4] = bc.addField(cpMyState, "ArrayA", "[Ljava/lang/Object;");
    // check subroutine
    for (int i = 0; i < invokes; i++) {
      opINVOKE invoke = (opINVOKE)modify.elementAt(i+1);
      if ( invoke.state.getSubroutine() != null )
	throw new CodeModifyException("CodeModify.Modify", "invoke in subroutine");
    }
    // start modifying
    for (int i = 0; i < invokes; i++) {
      ModifyINVOKE((opINVOKE)modify.elementAt(i+1), i);
      bc.addMigratory((opINVOKE)modify.elementAt(i+1));
    }
    ModifyStart();
    ModifyCatch();
    for (int i = 0; i < invokes; i++)
      bc.addException(la_start[i], la_end[i], la_catch[i], cpNotifyGone);
    bc.setMaxLocals(nmaxl);
    bc.setMaxStack(nmaxs);
  }

  public void ModifyStart() throws CodeModifyException {
    Instruction top = (Instruction)modify.elementAt(0);
    Label ltop = bc.lm.RegistLabel(top);
    Stack code = new Stack();
    int i;
    // reorder arguments
    Var locals[] = new Var[nmaxl];
    top.state.getNewLocals(locals, lstack);
    int arg_start = (bc.isStatic() ? 0 : 1);
    if ( arg_start != lmps )
      code.push(Var.vnull.load(arg_start));
    for (i = arg_start; i < maxl; i++) {
      if ( locals[i] == Var.vinv )
	break;
      code.push(locals[i].load(i+1));
      code.push(locals[i].store(i));
      if ( locals[i].isCategory2() )
	i++;
    }
    if ( arg_start != lmps )
      code.push(Var.vnull.store(lmps));
    // resume code
    int invokes = modify.size()-1;
    Vector resumes = new Vector();
    for (i = 0; i < invokes; i++) {
      opINVOKE invoke = (opINVOKE)modify.elementAt(i+1);
      Vector v = ModifyResume(invoke, i);
      la_resume[i] = bc.lm.RegistLabel((Instruction)v.elementAt(0));
      if ( invoke.offset == 0 )
        ltop = la_start[i];
      resumes.addElement(v);
    }
    // resume or not
    Instruction resume = Var.vnull.load(lmps);
    Label lresume = bc.lm.RegistLabel(resume);
    code.push(Var.vnull.load(lmps));
    code.push(new opCOND(IFNONNULL, lresume));
    code.push(new opCONST(ACONST_NULL));
    code.push(Var.vnull.store(lms));
    code.push(new opGOTO(GOTO, ltop));
    code.push(resume);
    code.push(new opFIELD(GETFIELD, cpStackFrameChild));
    code.push(new opCHECKCAST(CHECKCAST, cpMyState));
    code.push(Var.vnull.store(lms));
    if ( invokes > 0 ) {
      code.push(Var.vnull.load(lms));
      code.push(new opFIELD(GETFIELD, cpMyStateMEntryPoint));
      code.push(new opTABLE(TABLESWITCH, 0, invokes-1, ltop, la_resume));
      for (i = 0; i < resumes.size(); i++) {
	Vector v = (Vector)resumes.elementAt(i);
	for (int j = 0; j < v.size(); j++)
	  code.push(v.elementAt(j));
      }
    }
    // write
    bc.insertInsts(top.pc, code);
  }

  public Vector ModifyResume(opINVOKE invoke, int num) throws CodeModifyException {
    Vector code = new Vector();
    Var locals[] = new Var[nmaxl];
    invoke.state.getNewLocals(locals, lstack);
    if ( fDebug ) {
      System.out.println("\ninvoke #" + num);
      invoke.state.ShowDebug();
    }
    int idx[] = new int[5];
    for (int i = 0; i < nmaxl; i++) {
      if ( i > lstack && ++i >= nmaxl )
	break;
      if ( i == maxl ) i = lstack;
      if ( locals[i] == Var.vinv )
	continue;
      if ( locals[i] == Var.vnull ) {
	code.addElement(new opCONST(ACONST_NULL));
	code.addElement(Var.vnull.store(i));
	continue;
      }
      if ( locals[i].getVMType() > 4 )
	continue;
      ModifyResumeSub(code, locals[i], i, idx[locals[i].getVMType()]++);
    }
    code.addElement(new opGOTO(GOTO, la_start[num]));
    return code;
  }

  public void ModifyResumeSub(Vector code, Var v, int lidx, int aidx)
    throws CodeModifyException {
    if ( aidx == 0 ) {
      code.addElement(Var.vnull.load(lms));
      code.addElement(new opFIELD(GETFIELD, cpArray[v.getVMType()]));
      code.addElement(Var.vnull.store(ltmp+v.getVMType()));
    }
    code.addElement(Var.vnull.load(ltmp+v.getVMType()));
    code.addElement(new opPUSH(BIPUSH, aidx));
    code.addElement(v.aload());
    if ( v.getVMType() == 4 ) {
      int cindex = bc.addClass(v.getClassName());
      code.addElement(new opCHECKCAST(CHECKCAST, cindex));
    }
    code.addElement(v.store(lidx));
  }

  public void ModifyINVOKE(opINVOKE invoke, int num) throws CodeModifyException {
    Stack code = new Stack();
    // copy stack to local
    Var locals[] = new Var[nmaxl];
    invoke.state.getNewLocals(locals, lstack);
    int i, depth = 0;
    for (i = lstack; i < locals.length; i += 2) {
      if ( locals[i] == Var.vinv )
	break;
      code.push(locals[i].store(i)); depth++;
    }
    // back to stack
    int argc = bc.getMethodVars(invoke).size() - 1; // delete return value
    Vector back = new Vector();
    int lptr = lstack + (depth-1) * 2;
    for (i = 0; i < depth-argc; i++, lptr -= 2)
      back.addElement(locals[lptr].load(lptr));
    back.addElement(Var.vnull.load(lms));
    back.addElement(new opCONST(ACONST_NULL));
    back.addElement(Var.vnull.store(lms));
    for (i = 0; i < argc; i++, lptr -= 2)
      back.addElement(locals[lptr].load(lptr));
    for (i = 0; i < back.size(); i++)
      code.push(back.elementAt(i));
    // regist labels
    try {
      la_start[num] = bc.lm.RegistLabel((Instruction)back.elementAt(0));
      la_end[num]   = bc.lm.RegistLabel(bc.NextInst(invoke));
    } catch (DataFlowException e) {
      e.Display();
    }
    bc.insertInsts(invoke.pc, code);
  }

  public void ModifyCatch() throws CodeModifyException {
    Stack code = new Stack();
    int invokes = modify.size()-1;
    for (int i = 0; i < invokes; i++) {
      Vector v = ModifySave((opINVOKE)modify.elementAt(i+1), i);
      la_catch[i] = bc.lm.RegistLabel((Instruction)v.elementAt(0));
      for (int j = 0; j < v.size(); j++)
	code.push(v.elementAt(j));
    }
    bc.insertInsts(bc.getLastInstNumber(), code);
  }

  public Vector ModifySave(opINVOKE invoke, int num) throws CodeModifyException {
    Vector code = new Vector();
    code.addElement(Var.vnull.store(lnotify));
    code.addElement(new opNEW(NEW, cpMyState));
    code.addElement(new opDUP(DUP));
    code.addElement(new opINVOKE(INVOKESPECIAL, cpMyStateInit, 0));
    code.addElement(Var.vnull.store(lms));
    code.addElement(Var.vnull.load(lms));
    code.addElement(new opPUSH(BIPUSH, num));
    code.addElement(new opFIELD(PUTFIELD, cpMyStateMEntryPoint));
    // save here
    Var locals[] = new Var[nmaxl];
    invoke.state.getNewLocals(locals, lstack);
    int idx[] = new int[5];
    for (int i = 0; i < nmaxl; i++) {
      if ( i > lstack && ++i >= nmaxl )
	break;
      if ( i == maxl ) i = lstack;
      if ( locals[i] == Var.vinv || locals[i] == Var.vnull )
	continue;
      if ( locals[i].getVMType() > 4 )
	continue;
      ModifySaveSub(code, locals[i], i, idx[locals[i].getVMType()]++);
    }
    //
    code.addElement(Var.vnull.load(lms));
    code.addElement(Var.vnull.load(lnotify));
    code.addElement(new opFIELD(GETFIELD, cpNotifyGoneBottomofStack));
    code.addElement(new opFIELD(PUTFIELD, cpStackFrameChild));
    code.addElement(Var.vnull.load(lnotify));
    code.addElement(Var.vnull.load(lms));
    code.addElement(new opFIELD(PUTFIELD, cpNotifyGoneBottomofStack));
    code.addElement(Var.vnull.load(lnotify));
    code.addElement(new opATHROW(ATHROW));
    return code;
  }

  public void ModifySaveSub(Vector code, Var v, int lidx, int aidx)
    throws CodeModifyException {
    if ( aidx == 0 ) {
      code.addElement(Var.vnull.load(lms));
      code.addElement(new opFIELD(GETFIELD, cpArray[v.getVMType()]));
      code.addElement(Var.vnull.store(ltmp+v.getVMType()));
    }
    code.addElement(Var.vnull.load(ltmp+v.getVMType()));
    code.addElement(new opPUSH(BIPUSH, aidx));
    code.addElement(v.load(lidx));
    code.addElement(v.astore());
  }
}
