package DataFlow;
import java.util.*;
import ByteCode.*;
import ClassFile.*;
import Misc.*;

public class DataFlow
{
  private boolean fDebug;
  private ByteCode bc;

  public DataFlow(ByteCode bytecode, boolean fdebug) {
    bc = bytecode;
    fDebug = fdebug;
  }

  public void Analyze() throws DataFlowException {
    String here = "DataFlow.Analyze";
    State stat = new State(bc.getMaxStack(), bc.getMaxLocals());
    FlowManager fmn = new FlowManager(FlowManager.FM_NORMAL);
    FlowManager fme = new FlowManager(FlowManager.FM_EXCEPTION);
    FlowManager fmr = new FlowManager(FlowManager.FM_RET);
    Flow flow;
    NextState ns;
    Object obj;

    //
    // initializing
    //
    if ( fDebug )
      System.out.println("Analyzing starts: " + bc.getName() + "\n");
    bc.lm.MarkCheckPoints();	// mark points to save state
    try {			// set initial locals
      Vector v = bc.getInitialLocalVars();
      int i, j, sz = v.size();
      for (i = j = 0; i < sz; i++, j++) {
	Var var = (Var)v.elementAt(i);
	stat.local.Write(j, var);
	if ( var.isCategory2() ) j++;
      }
    } catch (MyException e) {
      e.Display();
      throw new DataFlowException(here, "initializing locals");
    }
    fmn.push(new Flow(bc.elementAt(0), stat));

    //
    // analyzing
    //
    while ( !fmn.empty() ) {
      flow = fmn.pop();
      if ( fDebug ) {
	fmn.ShowDebug(); fme.ShowDebug(); fmr.ShowDebug();
	System.out.println("offset: "+flow.inst.offset+", opcode: "+flow.inst.opcode);
	flow.stat.ShowDebug();
	System.out.println();
      }
      if ( (ns = flow.inst.Analyze(flow.stat)) != null ) {
	fme.push(bc.CatchException(flow.inst, ns.except));
	if ( ns.isRET )
	  while ( !ns.empty() ) fmr.push(ns.pop());
	else
	  while ( !ns.empty() ) fmn.push(ns.pop());
      }
      fmn.push(fmr.getRetStart());
      if ( fmn.empty() )
	while ( !fme.empty() ) fmn.push(fme.pop());
    }
    if ( !fme.empty() || !fmr.empty() )
      System.err.println("Unreachable code exists\n");
    if ( fDebug )
      System.out.println("Analyzing ends\n");
  }
}
