package DataFlow;
import java.util.*;
import Misc.*;

public class FlowManager
{
  private int type;
  private Stack flows;
  private Hashtable hash;

  public static final int FM_NORMAL=0, FM_EXCEPTION=1, FM_RET=2;

  public FlowManager(int t) {
    type = t;
    flows = new Stack();
    hash = new Hashtable();
  }

  public void push(Flow flow) throws DataFlowException {
    State stat = (State)hash.get(flow.inst);
    if ( stat != null ) {
      switch ( type ) {
      case FM_EXCEPTION: stat.MergeEX(flow.stat); break;
      case FM_RET:	 stat.MergeRET(flow.stat); break;
      default:		 stat.Merge(flow.stat); break;
      }
    } else {
      flows.push(flow);
      hash.put(flow.inst, flow.stat);
    }
  }

  public void push(MySet fset) throws DataFlowException {
    int i, sz = fset.size();
    for (i = sz-1; i >= 0; i--)
      push((Flow)fset.elementAt(i));
  }

  public Flow pop() throws DataFlowException {
    if ( flows.empty() )
      throw new DataFlowException("Flow.pop", null);
    Flow flow = (Flow)flows.pop();
    hash.remove(flow.inst);
    return flow;
  }

  public boolean empty() {
    return flows.empty();
  }

  public MySet getRetStart() throws DataFlowException {
    if ( type != FM_RET )
      throw new DataFlowException("FlowManager.getRetStart", null);
    MySet fset = new MySet();
    Stack stmp = new Stack();
    Hashtable htmp = new Hashtable();
    while ( !empty() ) {
      Flow flow = pop();
      if ( flow.inst.state != null )
	fset.add(flow);
      else {
	stmp.push(flow);
	htmp.put(flow.inst, flow.stat);
      }
    }
    flows = stmp;
    hash = htmp;
    return fset;
  }

  //
  // debug
  //
  public void ShowDebug() {
    System.out.print("flow" + type + ":");
    for (int i = flows.size()-1; i >= 0; i--)
      System.out.print(" " + ((Flow)flows.elementAt(i)).inst.offset);
    System.out.println();
  }
}
