package DataFlow;
import java.util.*;
import ByteCode.*;

public class State
{
  private StateCall call;  // subroutine call
  public StateStack stack; // operand stack
  public StateLocal local; // local variables
  public boolean isRET;

  public State(int maxs, int maxl) {
    stack = new StateStack(maxs);
    local = new StateLocal(maxl);
    call  = new StateCall();
  }

  public State(State ref) {
    stack = new StateStack(ref.stack);
    local = new StateLocal(ref.local);
    call  = new StateCall(ref.call);
    isRET = ref.isRET;
  }

  public void Enter(Label subr) throws DataFlowException {
    local.Enter();
    call.Enter(subr);
  }

  public Label Leave() throws DataFlowException {
    isRET = true;
    return call.Leave();
  }

  // normal&normal or ret&normal
  public boolean Merge(State ref) throws DataFlowException {
    if ( isRET ) {
      isRET = false;
      call = new StateCall(ref.call);
      return (stack.Merge(ref.stack) && local.MergeRET(ref.local));
    }
    return (stack.Merge(ref.stack) && local.Merge(ref.local) && call.Merge(ref.call, false));
  }

  // ex&ex
  public boolean MergeEX(State ref) throws DataFlowException {
    return (stack.Merge(ref.stack) && local.Merge(ref.local) && call.Merge(ref.call, true));
  }

  // ret&ret
  public boolean MergeRET(State ref) throws DataFlowException {
    return (stack.Merge(ref.stack) && local.Merge(ref.local));
  }

  //
  // CodeModify
  //
  public void getNewLocals(Var vars[], int sstart) {
    for (int i = 0; i < vars.length; i++)
      vars[i] = Var.vinv;
    local.getNewLocals(vars);
    stack.getNewLocals(vars, sstart);
  }

  public Label getSubroutine() {
    return call.getTop();
  }

  //
  // debug
  //
  public void ShowDebug() {
    stack.ShowDebug();
    local.ShowDebug();
    call.ShowDebug(isRET);
  }
}
