package DataFlow;
import java.util.*;
import ByteCode.*;

public class StateLocal implements DataFlowConst
{
  private int	 max;
  private Var	 locals[];
  private BitSet bits;

  public StateLocal(int maxl) {
    max = maxl;
    locals = new Var[max];
    for (int i = 0; i < max; i++)
      locals[i] = Var.vinv;
    bits = new BitSet(max);
  }

  public StateLocal(StateLocal ref) {
    max = ref.max;
    locals = new Var[max];
    for (int i = 0; i < max; i++)
      locals[i] = ref.locals[i];
    bits = (BitSet)ref.bits.clone();
  }

  //
  // operations
  //
  private static final String strIndex = "index exceeds";
  private static final String strInvalid  = "invalid access";
  private static final String strTypeMismatch  = "type mismatch";

  public Var Read(int index) throws DataFlowException {
    String here = "StateLocal.Read(int)";
    if ( index >= locals.length )
      throw new DataFlowException(here, strIndex);
    if ( locals[index].isCompatible(T_INVALID) )
      throw new DataFlowException(here, strInvalid);
    return locals[index];
  }

  public Var Read(int index, int vt) throws DataFlowException {
    String here = "StateLocal.Read(int, int)";
    Var v = Read(index);
    if ( !v.isCompatible(vt) )
      throw new DataFlowException(here, strTypeMismatch);
    return v;
  }

  public void Write(int index, Var v) throws DataFlowException {
    String here = "StateLocal.Write(int)";
    if ( index >= max || v == Var.vinv )
      throw new DataFlowException(here, null);
    if ( index-1 >= 0 && locals[index-1].isCategory2() )
      locals[index-1] = Var.vinv;
    if ( v.isCategory2() ) {
      if ( index+1 >= max )
	throw new DataFlowException(here, strIndex);
      if ( index+1 < max && locals[index+1].isCategory2() )
	locals[index+2] = Var.vinv;
      locals[index+1] = Var.vinv;
      bits.set(index+1);
    }
    locals[index] = v;
    bits.set(index);
  }

  public void Enter() throws DataFlowException {
    bits = new BitSet();
  }

  //
  // merge
  //
  public boolean Merge(StateLocal ref) throws DataFlowException {
    DataFlowException e = new DataFlowException("StateLocal.Merge", null);
    boolean funchanged = true;
    if ( max != ref.max )
      throw e;
    bits.or(ref.bits);
    for (int i = 0; i < max; i++) {
      Var v1 = locals[i];
      Var v2 = ref.locals[i];
      Var v3 = v1.Merge(v2, false);
      if ( !v1.equals(v3) || !v2.equals(v3) ) {
	locals[i] = v3;
	funchanged = false;
      }
    }
    return funchanged;
  }

  public boolean MergeRET(StateLocal ref) throws DataFlowException {
    DataFlowException e = new DataFlowException("StateLocal.MergeRET", null);
    boolean funchanged = true;
    if ( max != ref.max )
      throw e;
    for (int i = 0; i < max; i++) {
      if ( bits.get(i) ) {
	if ( !locals[i].equals(ref.locals[i]) )
	  funchanged = false;
	if ( i-1 >= 0 && !bits.get(i) && locals[i-1].isCategory2() )
	  locals[i-1] = Var.vinv;
      } else
	locals[i] = ref.locals[i];
    }
    bits.or(ref.bits);
    return funchanged;
  }

  //
  // CodeModify
  //
  public void getNewLocals(Var vars[]) {
    for (int i = 0; i < max; i++)
      vars[i] = locals[i];
  }

  //
  // debug
  //
  public void ShowDebug() {
    System.out.print("local:");
    for (int i = 0; i < max; i++)
      System.out.print(" " + i + (bits.get(i) ? "=" : "?") + locals[i].ShowDebug());
    System.out.print("\n");
  }
}
