package DataFlow;
import java.util.*;

public class StateStack implements DataFlowConst
{
  private int max;
  private Stack stack;

  public StateStack(int maxs) {
    max = maxs;
    stack = new Stack();
  }

  public StateStack(StateStack ref) {
    int i, sz = ref.stack.size();
    max = ref.max;
    stack = new Stack();
    for (i = 0; i < sz; i++)
      stack.addElement(ref.stack.elementAt(i));
  }

  //
  // operations
  //
  private static String strStackEmpty    = "stack empty";
  private static String strTypeMismatch  = "type mismatch";
  private static String strStackOverflow = "stack overflow";

  public Var pop() throws DataFlowException {
    String here = "StateStack.pop()";
    if ( stack.empty() )
      throw new DataFlowException(here, strStackEmpty);
    return (Var)stack.pop();
  }

  public Var pop(int vt) throws DataFlowException {
    String here = "StateStack.pop(int)";
    Var top = (Var)stack.pop();
    if ( !top.isCompatible(vt) )
      throw new DataFlowException(here, strTypeMismatch);
    return top;
  }

  public Var pop(Var v) throws DataFlowException {
    String here = "StateStack.pop(Var)";
    Var top = (Var)stack.pop();
    if ( !top.isCompatible(v) )
      throw new DataFlowException(here, strTypeMismatch);
    return top;
  }

  public void push(Var v) throws DataFlowException {
    String here = "StateStack.push";
    stack.push(v);
    if ( stack.size() > max )
      throw new DataFlowException(here, strStackOverflow);
  }

  public Var peek(int vt) throws DataFlowException {
    String here = "StateStack.peek(Var)";
    if ( stack.empty() )
      throw new DataFlowException(here, strStackEmpty);
    Var top = (Var)stack.peek();
    if ( !top.isCompatible(vt) )
      throw new DataFlowException(here, strTypeMismatch);
    return top;
  }

  public void Flash() {
    stack = new Stack();
  }

  //
  // merge
  //
  public boolean Merge(StateStack s) throws DataFlowException {
    DataFlowException e = new DataFlowException("StateStack.Merge", null);
    boolean funchanged = true;
    int i, sz = stack.size();
    if ( sz != s.stack.size() || max != s.max )
      throw e;
    for (i = 0; i < sz; i++) {
      Var v1 = (Var)stack.elementAt(i);
      Var v2 = (Var)s.stack.elementAt(i);
      Var v3 = v1.Merge(v2, true);
      if ( !v1.equals(v3) ) {
	stack.setElementAt(v3, i);
	funchanged = false;
      }
    }
    return funchanged;
  }

  //
  // CodeModify
  //
  public void getNewLocals(Var vars[], int start) {
    int i, sz = stack.size() - 1;
    for (i = sz; i >= 0; i--)
      vars[start+(sz-i)*2] = (Var)stack.elementAt(i);
  }

  //
  // debug
  //
  public void ShowDebug() {
    System.out.print("stack:");
    for (int i = stack.size()-1; i >= 0; i--)
      System.out.print(" " + ((Var)stack.elementAt(i)).ShowDebug());
    System.out.print("\n");
  }
}
