/**
 * JavaGo Translator - jgox
 * 
 * 1999-2000 Takahiro SAKAMOTO / takas@yl.is.s.u-tokyo.ac.jp
 */
import java.io.*;
import ByteCode.*;
import ClassFile.*;
import CodeModify.*;
import DataFlow.*;
import Misc.*;

public class jgox implements ClassFileConst
{
  public static boolean fDebugCodeModify;
  public static boolean fDebugDataFlow;

  public static void main(String[] args) throws Exception {
    //
    // default debug options
    //
    fDebugCodeModify = false;
    fDebugDataFlow   = false;

    //
    // parse command line
    //
    if ( args.length < 2 ) {
      System.out.println("Usage: jgox <.class> <method>");
      return;
    }
    String src = args[0], dst, ext = ".class", mname = args[1];
    if ( !src.endsWith(ext) ) src += ext;
    dst = src + ".out";
    System.out.println("JavaGO bytecode translator - jgox -\n" +
		       "input  class: " + src + "\n" +
		       "output class: " + dst + "\n" +
		       " method name: " + mname + "\n");

    //
    // action
    //
    try {
      FileInputStream  srcFileS = new FileInputStream(src);
      DataInputStream  srcDataS = new DataInputStream(srcFileS);
      FileOutputStream dstFileS = new FileOutputStream(dst);
      DataOutputStream dstDataS = new DataOutputStream(dstFileS);
      try {
	ClassFile  cf = new ClassFile(srcDataS);
	ByteCode   bc;
	CodeModify cm;
	DataFlow   df;
	for (int i = 0; i < cf.methods.length; i++) {
	  if ( (cf.methods[i].flag & ACC_ABSTRACT) != 0 )
	    continue;
	  if ( !cf.methods[i].name.cp.getString().equals(mname) )
	    continue;
	  bc = new ByteCode(cf.methods[i]);
	  cm = new CodeModify(bc, fDebugCodeModify);
	  df = new DataFlow(bc, fDebugDataFlow);
	  if ( !cm.Survey() ) {
	    System.out.println("This method can't be translated: " + bc.getName());
	    continue;
	  }
	  long millis = System.currentTimeMillis();
	  df.Analyze();
	  long analyze = System.currentTimeMillis();
	  cm.Modify();
	  long modify = System.currentTimeMillis();
	  bc.Commit();
	  System.out.println("analysis: " + (analyze - millis)
			     + "\nmodify  : " + (modify - analyze)
			     + "\ntotal   : " + (modify - millis));
	}
	cf.Save(dstDataS);
	System.out.println("Successfully finished.");
      } finally {
	srcFileS.close();
	dstFileS.close();
      }
    } catch (MyException e) {
      e.Display();
    //} catch (Exception e) {
    //  System.out.println("jgox: " + e.toString());
    }
  }
}
